import base64
from cryptography.hazmat.primitives.asymmetric import ed25519
from cryptography.hazmat.primitives import serialization
import hashlib

class Onion:

    def __init__(self):
        self.key_type = 'NEW'
        self.key = 'BEST'
        self.id = None
        self.ports = {}

    @classmethod
    def random(cls):
        ''' generate new random Onion object '''
        private_key = ed25519.Ed25519PrivateKey.generate()
        return cls.from_key(private_key)

    @classmethod
    def from_key(cls, private_key):
        ''' create Onion object from existing ed25519 key '''
        onion = cls()
        onion.key_type = 'ED25519-V3'
        onion.key = format_key(private_key)
        onion.id = calculate_id(private_key.public_key())
        return onion


def format_key(private_key):
    ''' convert ed25519 key to Tor format '''
    b = private_key.private_bytes(
        encoding=serialization.Encoding.Raw,
        format=serialization.PrivateFormat.Raw,
        encryption_algorithm=serialization.NoEncryption(),
    )
    h = hashlib.sha512()
    h.update(b[:32])
    p = list(h.digest())
    p[0] &= 248
    p[31] &= 127
    p[31] |= 64
    return base64.b64encode(bytes(p)).decode('utf8')

def calculate_id(public_key):
    ''' calculate onion service ID from ed25519 public key '''
    b = public_key.public_bytes(
        encoding=serialization.Encoding.Raw,
        format=serialization.PublicFormat.Raw,
    )
    h = hashlib.sha3_256()
    h.update(b'.onion checksum')
    h.update(b)
    h.update(b'\x03')
    d = h.digest()
    c = d[:2]
    combined = b + c + b'\x03'
    service_id = base64.b32encode(combined)
    return service_id.decode('utf8').replace('=', '').lower()




import asyncio

class Events:

    def __init__(self, controller):
        self.controller = controller
        self.queue = asyncio.Queue()
        self.__listeners = {}
        self.__events = set()
        self.task = asyncio.create_task(self.__loop())

    async def __loop(self):
        ''' main event dispatch loop '''
        while True:
            event = await self.queue.get()
            await self.__dispatch(event)
            self.queue.task_done()

    async def __update_events(self):
        ''' send SETEVENTS command when set of event listeners changes '''
        events = set(self.__listeners.keys())
        if events != self.__events:
            self.__events = events
            eventstr = ' '.join(events)
            resp = await self.controller.io.cmd('SETEVENTS ' + eventstr)
            if resp['status'] != 250:
                raise Exception('unable to send SETEVENTS')

    async def on(self, event, listener):
        ''' add listener for event type '''
        if event not in self.__listeners:
            self.__listeners[event] = set()
        self.__listeners[event].add(listener)
        await self.__update_events()

    async def off(self, event, listener):
        ''' remove listener for event type '''
        if event in self.__listeners:
            self.__listeners[event].discard(listener)
            if len(self.__listeners[event]) == 0:
                del self.__listeners[event]
        await self.__update_events()

    async def __dispatch(self, event):
        ''' dispatch event '''
        if event.type in self.__listeners:
            for listener in self.__listeners[event.type]:
                await listener(event)


class Event:
    pass


class BandwidthEvent(Event):

    type = 'BW'

    def __init__(self, read, written, **kwargs):
        self.read = read
        self.written = written
        self.kwargs = kwargs


class CircuitEvent(Event):

    type = 'CIRC'

    def __init__(self, id, status, path=None, **kwargs):
        self.id = id
        self.status = status
        self.path = path
        self.kwargs = kwargs


class StreamEvent(Event):

    type = 'STREAM'

    def __init__(self, id, status, circ_id, target, **kwargs):
        self.id = id
        self.status = status
        self.circ_id = circ_id
        self.target = target
        self.kwargs = kwargs


class AddrMapEvent(Event):

    type = 'ADDRMAP'

    def __init__(self, hostname, destination, expiry, **kwargs):
        self.hostname = hostname
        self.destination = destination
        self.expiry = expiry
        self.kwargs = kwargs


class HiddenServiceEvent(Event):

    type = 'HS_DESC'

    def __init__(self, *args, **kwargs):
        self.action = args[0]
        self.address = args[1]
        self.authentication = args[2]
        self.directory = args[3]
        if len(args) > 4:
            self.descriptor_id = args[4]
        else:
            self.descriptor_id = None
        self.kwargs = kwargs


class StreamBandwidthEvent(Event):

    type = 'STREAM_BW'

    def __init__(self, id, written, read, time, **kwargs):
        self.id = id
        self.written = written
        self.read = read
        self.time = time
        self.kwargs = kwargs


class NetworkLivenessEvent(Event):

    type = 'NETWORK_LIVENESS'

    def __init__(self, status, **kwargs):
        self.status = status
        self.kwargs = kwargs


class GuardEvent(Event):

    type = 'GUARD'

    def __init__(self, guard_type, endpoint, status, **kwargs):
        self.guard_type = guard_type
        self.endpoint = endpoint
        self.status = status
        self.kwargs = kwargs


class SignalEvent(Event):

    type = 'SIGNAL'

    def __init__(self, signal, **kwargs):
        self.signal = signal
        self.kwargs = kwargs


class OrConnEvent(Event):

    type = 'ORCONN'

    def __init__(self, endpoint, status, **kwargs):
        self.endpoint = endpoint
        self.status = status
        self.kwargs = kwargs


class CircMinorEvent(Event):

    type = 'CIRC_MINOR'

    def __init__(self, id, event, path=None, **kwargs):
        self.id = id
        self.event = event
        self.path = path
        self.kwargs = kwargs


class StatusGeneralEvent(Event):

    type = 'STATUS_GENERAL'

    def __init__(self, runlevel, action, **kwargs):
        self.runlevel = runlevel
        self.action = action
        self.kwargs = kwargs


class StatusClientEvent(Event):

    type = 'STATUS_CLIENT'

    def __init__(self, runlevel, action, **kwargs):
        self.runlevel = runlevel
        self.action = action
        self.kwargs = kwargs


class StatusServerEvent(Event):

    type = 'STATUS_SERVER'

    def __init__(self, runlevel, action, **kwargs):
        self.runlevel = runlevel
        self.action = action
        self.kwargs = kwargs


class HSDescContentEvent(Event):

    type = 'HS_DESC_CONTENT'

    def __init__(self, address, descriptor_id, directory, **kwargs):
        self.address = address
        self.descriptor_id = descriptor_id
        self.directory = directory
        self.kwargs = kwargs


class TransportLaunchedEvent(Event):

    type = 'TRANSPORT_LAUNCHED'

    def __init__(self, transport_type, name, address, port, **kwargs):
        self.transport_type = transport_type
        self.name = name
        self.address = address
        self.port = port
        self.kwargs = kwargs


# registered event types
EVENT_TYPES = {
    'BW': BandwidthEvent,
    'CIRC': CircuitEvent,
    'STREAM': StreamEvent,
    'ADDRMAP': AddrMapEvent,
    'HS_DESC': HiddenServiceEvent,
    'STREAM_BW': StreamBandwidthEvent,
    'NETWORK_LIVENESS': NetworkLivenessEvent,
    'GUARD': GuardEvent,
    'SIGNAL': SignalEvent,
    'ORCONN': OrConnEvent,
    'CIRC_MINOR': CircMinorEvent,
    'STATUS_GENERAL': StatusGeneralEvent,
    'STATUS_CLIENT': StatusClientEvent,
    'STATUS_SERVER': StatusServerEvent,
    'HS_DESC_CONTENT': HSDescContentEvent,
    'TRANSPORT_LAUNCHED': TransportLaunchedEvent,
}



import asyncio
from .events import EVENT_TYPES

class Parser:

    def __init__(self):
        self.index = 0
        self.text = ''
        self.key = ''
        self.value = ''
        self.args = []
        self.kwargs = {}

    def parse(self, text):
        ''' Parse text for args and kwargs '''
        self.index = 0
        self.text = text
        while self.index < len(self.text):
            c = self.pop()
            if c == ' ':
                self._flush()
                continue
            elif c == '"':
                if not self.value:
                    self._parse_quoted()
                    self._flush()
                    continue
            elif c == '=':
                if not self.key:
                    self.key = self.value
                    self.value = ''
                    continue
            elif c == '+':
                if not self.value:
                    a = self.text.index('=\r\n', self.index)
                    key = self.text[self.index:a]
                    a += 3
                    b = self.text.index('.\r\n', a)
                    value = self.text[a:b]
                    b += 3
                    self.index = b
                    self.kwargs[key] = value
                    continue
            self.value += c
        self._flush()

    def parse_keywords(self, text):
        ''' Parse only keyword response text (used in getinfo responses) '''
        self.index = 0
        self.text = text
        while self.index < len(self.text):
            c = self.pop()
            if c == '"':
                if not self.value:
                    self._parse_quoted()
                    self._flush()
                    continue
            elif c == '=':
                if not self.key:
                    self.key = self.value
                    try:
                        a = self.text.index('\r\n', self.index)
                    except ValueError:
                        a = -1
                    if a > -1:
                        self.value = self.text[self.index:a]
                        self._flush()
                        self.index = a + 3
                        continue
                    else:
                        self.value = self.text[self.index:]
                        self._flush()
                        break
            elif c == '+':
                if not self.value:
                    a = self.text.index('=\r\n', self.index)
                    key = self.text[self.index:a]
                    a += 3
                    b = self.text.index('\r\n.\r\n', a)
                    value = self.text[a:b]
                    b += 3
                    self.index = b
                    self.kwargs[key] = value
                    continue
            self.value += c
        self._flush()
        if self.key:
            self.kwargs[self.key] = ''

    def _flush(self):
        if self.value:
            if self.key:
                self.kwargs[self.key] = self.value
                self.key = ''
            else:
                self.args.append(self.value)
            self.value = ''

    def _parse_quoted(self):
        while True:
            c = self.pop()
            if c == '\\':
                self.value += self.pop()
            elif c == '"':
                return
            else:
                self.value += c

    def pop(self):
        c = self.text[self.index]
        self.index += 1
        return c


def parse(text):
    ''' parse text and return args, kwargs '''
    p = Parser()
    p.parse(text)
    return p.args, p.kwargs

def parse_keywords(text):
    ''' parse text and return only kwargs '''
    p = Parser()
    p.parse_keywords(text)
    return p.kwargs


class TextProtocol:

    def __init__(self, r, w, event_queue=None):
        self.r = r
        self.w = w
        self.response_queue = asyncio.Queue()
        self.event_queue = event_queue
        self.lock = asyncio.Lock()
        self.task = asyncio.create_task(self.__loop())

    async def __loop(self):
        ''' main reader loop '''
        while True:
            resp = await self.__read()
            if resp['status'] == 650:
                await self.__put_event(resp)
            else:
                await self.response_queue.put(resp)

    async def __put_event(self, resp):
        ''' parse and queue an event from a resp object '''
        if self.event_queue is None:
            return
        args, kwargs = parse(' '.join(resp['lines']))
        type, args = args[0], args[1:]
        if type in EVENT_TYPES:
            event = EVENT_TYPES[type](*args, **kwargs)
            await self.event_queue.put(event)

    async def __write(self, cmd):
        ''' writes a command '''
        self.w.write(cmd.encode('utf8') + b'\r\n')
        await self.w.drain()

    async def __read(self):
        ''' reads a response '''
        lines = []
        status = -1
        while True:
            line = await self.r.readline()
            # make sure line starts with valid status code
            try:
                status = int(line[:3])
            except ValueError:
                status = -1
                break
            # don't append trailing OK line
            if line[3:4] == b' ' and line[4:] == b'OK\r\n':
                break
            # + multiline response
            if line[3:4] == b'+':
                data = line[3:]
                while True:
                    line = await self.r.readline()
                    data += line
                    if line == b'.\r\n':
                        break
                lines.append(data.decode('utf8'))
                continue
            lines.append(line[4:].decode('utf8').strip())
            # if not multiline response, exit
            if line[3:4] != b'-':
                break
        return {'status': status, 'lines': lines}

    async def cmd(self, cmd):
        ''' send a command and return response object '''
        async with self.lock:
            await self.__write(cmd)
            resp = await self.response_queue.get()
            self.response_queue.task_done()
            return resp



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='python-aiotor',
    version='0.2.1',
    author='Davy Wybiral',
    author_email="davy.wybiral@gmail.com",
    description='asyncio Tor controller client',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/wybiral/aiotor",
    packages=['aiotor'],
    install_requires=['cryptography'],
    license='MIT',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import asyncio
import hashlib
import hmac
from os import urandom
from . import events
from .onions import Onion
from .textprotocol import parse, parse_keywords, TextProtocol

class Controller:

    def __init__(self, host='127.0.0.1', port=9051):
        self.host = host
        self.port = port
        self.events = events.Events(self)
        self.io = None
        self.auth = {
            'methods': [],
            'cookiefile': None
        }

    async def connect(self):
        ''' connect to tor controller '''
        r, w = await asyncio.open_connection(self.host, self.port)
        self.io = TextProtocol(r, w, event_queue=self.events.queue)
        self.__parse_protocolinfo(await self.io.cmd('PROTOCOLINFO 1'))

    def __parse_protocolinfo(self, resp):
        if resp['status'] != 250:
            raise Exception('Unable to connect')
        for line in resp['lines']:
            args, kwargs = parse(line)
            if args[0] == 'AUTH':
                self.auth['methods'] = kwargs['METHODS'].split(',')
                self.auth['cookiefile'] = kwargs.get('COOKIEFILE', None)

    async def authenticate(self, password=None):
        ''' authenticate using any available method '''
        methods = self.auth['methods']
        if 'NULL' in methods:
            resp = await self.__authenticate_none()
        elif 'HASHEDPASSWORD' and password is not None:
            resp = await self.__authenticate_password(password)
        elif 'SAFECOOKIE' in self.auth['methods'] and self.auth['cookiefile']:
            resp = await self.__authenticate_safecookie()
        elif 'COOKIE' in self.auth['methods'] and self.auth['cookiefile']:
            resp = await self.__authenticate_cookie()
        else:
            raise Exception('no authentication method available')
        if resp['status'] != 250:
            raise Exception('authentication failed')

    async def __authenticate_none(self):
        return await self.io.cmd('AUTHENTICATE')

    async def __authenticate_password(self, password):
        quoted = '"' + password + '"'
        return await self.io.cmd('AUTHENTICATE ' + quoted)

    async def __authenticate_safecookie(self):
        cookie = open(self.auth['cookiefile'], 'rb').read()
        # send client nonce
        client_nonce = urandom(32)
        challenge = 'AUTHCHALLENGE SAFECOOKIE ' + client_nonce.hex()
        resp = await self.io.cmd(challenge)
        args, kwargs = parse(resp['lines'][0])
        # authenticate server hash
        server_hash = bytes.fromhex(kwargs['SERVERHASH'])
        server_nonce = bytes.fromhex(kwargs['SERVERNONCE'])
        key = b'Tor safe cookie authentication server-to-controller hash'
        msg = cookie + client_nonce + server_nonce
        h = hmac.new(key, msg, hashlib.sha256).digest()
        if h != server_hash:
            raise Exception('invalid server hash')
        # construct client hash
        key = b'Tor safe cookie authentication controller-to-server hash'
        msg = cookie + client_nonce + server_nonce
        h = hmac.new(key, msg, hashlib.sha256).hexdigest()
        return await self.io.cmd('AUTHENTICATE ' + h)

    async def __authenticate_cookie(self):
        cookie = open(self.auth['cookiefile'], 'rb').read()
        return await self.io.cmd('AUTHENTICATE ' + cookie.hex())

    async def getinfo(self, key):
        resp = await self.io.cmd('GETINFO ' + key)
        if resp['status'] != 250:
            raise Exception('Request failed')
        text = ' '.join(resp['lines'])
        return parse_keywords(text)

    async def signal(self, signal):
        resp = await self.io.cmd('SIGNAL ' + signal)
        print(resp)
        if resp['status'] != 250:
            raise Exception('Request failed')

    async def add_onion(self, onion, wait=False):
        key_str = '{}:{}'.format(onion.key_type, onion.key)
        ports = onion.ports
        ports_str = ' '.join('Port={},{}'.format(k, ports[k]) for k in ports)
        resp = await self.io.cmd('ADD_ONION ' + key_str + ' ' + ports_str)
        if resp['status'] != 250:
            raise Exception('Request failed')
        args, kwargs = parse(' '.join(resp['lines']))
        onion.id = kwargs['ServiceID']
        if 'PrivateKey' in kwargs:
            key_type, key = kwargs['PrivateKey'].split(':', maxsplit=1)
            onion.key_type = key_type
            onion.key = key
        if wait:
            event = asyncio.Event()
            async def hs_desc(e):
                if e.address != onion.id:
                    return
                if e.action == 'UPLOADED':
                    event.set()
            await self.events.on('HS_DESC', hs_desc)
            await event.wait()
            await self.events.off('HS_DESC', hs_desc)
        return onion

    async def del_onion(self, onion):
        resp = await self.io.cmd('DEL_ONION ' + onion.id)
        if resp['status'] != 250:
            raise Exception('Request failed')